@@ -1,5 +1,5 @@
 /* JEditorPane.java --
-   Copyright (C) 2002, 2004, 2005  Free Software Foundation, Inc.
+   Copyright (C) 2002, 2004, 2005, 2006,  Free Software Foundation, Inc.
 
 This file is part of GNU Classpath.
 
@@ -38,13 +38,19 @@ exception statement from your version. */
 
 package javax.swing;
 
+import java.awt.Container;
 import java.awt.Dimension;
+import java.io.BufferedInputStream;
+import java.io.FilterInputStream;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.InputStreamReader;
 import java.io.Reader;
+import java.io.StringReader;
 import java.net.MalformedURLException;
 import java.net.URL;
+import java.net.URLConnection;
+import java.util.HashMap;
 
 import javax.accessibility.AccessibleContext;
 import javax.accessibility.AccessibleHyperlink;
@@ -53,6 +59,8 @@ import javax.accessibility.AccessibleStateSet;
 import javax.accessibility.AccessibleText;
 import javax.swing.event.HyperlinkEvent;
 import javax.swing.event.HyperlinkListener;
+import javax.swing.plaf.TextUI;
+import javax.swing.text.AbstractDocument;
 import javax.swing.text.BadLocationException;
 import javax.swing.text.DefaultEditorKit;
 import javax.swing.text.Document;
@@ -88,6 +96,7 @@ import javax.swing.text.html.HTMLEditorKit;
  *
  * @author original author unknown
  * @author Roman Kennke (roman@kennke.org)
+ * @author Anthony Balkissoon abalkiss at redhat dot com
  */
 public class JEditorPane extends JTextComponent
 {
@@ -167,6 +176,14 @@ public class JEditorPane extends JTextComponent
     extends AccessibleJEditorPane implements AccessibleHypertext
   {
 
+    /**
+     * Creates a new JEditorPaneAccessibleHypertextSupport object.
+     */
+    public JEditorPaneAccessibleHypertextSupport()
+    {
+      super();
+    }
+    
     /**
      * The accessible representation of a HTML link. 
      *
@@ -469,6 +486,34 @@ public class JEditorPane extends JTextComponent
     }
   }
 
+  /**
+   * Used to store a mapping for content-type to editor kit class.
+   */
+  private static class EditorKitMapping
+  {
+    /**
+     * The classname of the editor kit.
+     */
+    String className;
+
+    /**
+     * The classloader with which the kit is to be loaded.
+     */
+    ClassLoader classLoader;
+
+    /**
+     * Creates a new EditorKitMapping object.
+     * 
+     * @param cn the classname
+     * @param cl the classloader
+     */
+    EditorKitMapping(String cn, ClassLoader cl)
+    {
+      className = cn;
+      classLoader = cl;
+    }
+  }
+
   /**
    * An EditorKit used for plain text. This is the default editor kit for
    * JEditorPanes.
@@ -493,15 +538,162 @@ public class JEditorPane extends JTextComponent
     }
   }
 
+  /**
+   * A special stream that can be cancelled.
+   */
+  private class PageStream
+    extends FilterInputStream
+  {
+    /**
+     * True when the stream has been cancelled, false otherwise.
+     */
+    private boolean cancelled;
+
+    protected PageStream(InputStream in)
+    {
+      super(in);
+      cancelled = false;
+    }
+
+    private void checkCancelled()
+      throws IOException
+    {
+      if (cancelled)
+        throw new IOException("Stream has been cancelled");
+    }
+
+    void cancel()
+    {
+      cancelled = true;
+    }
+
+    public int read()
+      throws IOException
+    {
+      checkCancelled();
+      return super.read();
+    }
+
+    public int read(byte[] b, int off, int len)
+      throws IOException
+    {
+      checkCancelled();
+      return super.read(b, off, len);
+    }
+
+    public long skip(long n)
+      throws IOException
+    {
+      checkCancelled();
+      return super.skip(n);
+    }
+
+    public int available()
+      throws IOException
+    {
+      checkCancelled();
+      return super.available();
+    }
+
+    public void reset()
+      throws IOException
+    {
+      checkCancelled();
+      super.reset();
+    }
+  }
+
+  /**
+   * The thread that loads documents asynchronously.
+   */
+  private class PageLoader
+    implements Runnable
+  {
+    private Document doc;
+    private PageStream in;
+    private URL old;
+    URL page;
+    PageLoader(Document doc, InputStream in, URL old, URL page)
+    {
+      this.doc = doc;
+      this.in = new PageStream(in);
+      this.old = old;
+      this.page = page;
+    }
+
+    public void run()
+    {
+      try
+        {
+          read(in, doc);
+        }
+      catch (IOException ex)
+        {
+          UIManager.getLookAndFeel().provideErrorFeedback(JEditorPane.this);
+        }
+      finally
+        {
+          if (SwingUtilities.isEventDispatchThread())
+            firePropertyChange("page", old, page);
+          else
+            {
+              SwingUtilities.invokeLater(new Runnable()
+              {
+                public void run()
+                {
+                  firePropertyChange("page", old, page);
+                }
+              });
+            }
+         }
+     }
+
+     void cancel()
+     {
+       in.cancel();
+     }
+  }
+
   private static final long serialVersionUID = 3140472492599046285L;
   
-  private URL page;
   private EditorKit editorKit;
   
   boolean focus_root;
+  
+  /**
+   * Maps content-types to editor kit instances.
+   */
+  static HashMap editorKits;
+
+  // A mapping between content types and registered EditorKit types
+  static HashMap registerMap;
+
+  static
+  {
+    registerMap = new HashMap();
+    editorKits = new HashMap();
+    registerEditorKitForContentType("application/rtf",
+                                    "javax.swing.text.rtf.RTFEditorKit");
+    registerEditorKitForContentType("text/plain",
+                                    "javax.swing.JEditorPane$PlainEditorKit");
+    registerEditorKitForContentType("text/html",
+                                    "javax.swing.text.html.HTMLEditorKit");
+    registerEditorKitForContentType("text/rtf",
+                                    "javax.swing.text.rtf.RTFEditorKit");
+
+  }
+
+  // A mapping between content types and used EditorKits
+  HashMap editorMap;  
+
+  /**
+   * The currently loading stream, if any.
+   */
+  private PageLoader loader;
 
   public JEditorPane()
   {
+    init();
     setEditorKit(createDefaultEditorKit());
   }
 
@@ -512,24 +704,68 @@ public class JEditorPane extends JTextComponent
 
   public JEditorPane(String type, String text)
   {
+    init();
     setEditorKit(createEditorKitForContentType(type));
     setText(text);
   }
 
   public JEditorPane(URL url) throws IOException
   {
-    this();
+    init();
+    setEditorKit(createEditorKitForContentType("text/html"));
     setPage(url);
   }
+  
+  /**
+   * Called by the constructors to set up the default bindings for content 
+   * types and EditorKits.
+   */
+  void init()
+  {
+    editorMap = new HashMap();
+  }
 
   protected EditorKit createDefaultEditorKit()
   {
     return new PlainEditorKit();
   }
 
+  /**
+   * Creates and returns an EditorKit that is appropriate for the given 
+   * content type.  This is created using the default recognized types
+   * plus any EditorKit types that have been registered.
+   * 
+   * @see #registerEditorKitForContentType(String, String)
+   * @see #registerEditorKitForContentType(String, String, ClassLoader)
+   * @param type the content type
+   * @return an EditorKit for use with the given content type
+   */
   public static EditorKit createEditorKitForContentType(String type)
   {
-    return new PlainEditorKit();
+    // Try cached instance.
+    EditorKit e = (EditorKit) editorKits.get(type);
+    if (e == null)
+      {
+        EditorKitMapping m = (EditorKitMapping) registerMap.get(type);
+        if (m != null)
+          {
+            String className = m.className;
+            ClassLoader loader = m.classLoader;
+            try
+              {
+                e = (EditorKit) loader.loadClass(className).newInstance();
+              }
+            catch (Exception e2)
+              {    
+                // The reference implementation returns null when class is not
+                // loadable or instantiatable.
+              }
+          }
+        // Cache this for later retrieval.
+        if (e != null)
+          editorKits.put(type, e);
+      }
+    return e;
   }
 
   /**
@@ -578,49 +814,144 @@ public class JEditorPane extends JTextComponent
     return editorKit;
   }
 
+  /**
+   * Returns the class name of the EditorKit associated with the given
+   * content type.
+   * 
+   * @since 1.3
+   * @param type the content type
+   * @return the class name of the EditorKit associated with this content type
+   */
   public static String getEditorKitClassNameForContentType(String type)
   {
-    return "text/plain";
-  }
-
-  public EditorKit getEditorKitForContentType(String type)
-  {
-    return editorKit;
+    EditorKitMapping m = (EditorKitMapping) registerMap.get(type);
+    String kitName = m != null ? m.className : null;
+    return kitName;
   }
 
   /**
-   * Returns the preferred size for the JEditorPane.  
+   * Returns the EditorKit to use for the given content type.  If an
+   * EditorKit has been explicitly set via 
+   * <code>setEditorKitForContentType</code>
+   * then it will be returned.  Otherwise an attempt will be made to create
+   * an EditorKit from the default recognzied content types or any
+   * EditorKits that have been registered.  If none can be created, a
+   * PlainEditorKit is created.
+   * 
+   * @see #registerEditorKitForContentType(String, String)
+   * @see #registerEditorKitForContentType(String, String, ClassLoader)
+   * @param type the content type
+   * @return an appropriate EditorKit for the given content type
+   */
+  public EditorKit getEditorKitForContentType(String type)
+  {
+    // First check if an EditorKit has been explicitly set.
+    EditorKit e = (EditorKit) editorMap.get(type);
+    // Then check to see if we can create one.
+    if (e == null)
+      {
+        e = createEditorKitForContentType(type);
+        if (e != null)
+          setEditorKitForContentType(type, e);
+      }
+    // Otherwise default to PlainEditorKit.
+    if (e == null)
+      e = createDefaultEditorKit();
+    return e;
+  }
+
+  /**
+   * Returns the preferred size for the JEditorPane. This is implemented to
+   * return the super's preferred size, unless one of
+   * {@link #getScrollableTracksViewportHeight()} or
+   * {@link #getScrollableTracksViewportWidth()} returns <code>true</code>,
+   * in which case the preferred width and/or height is replaced by the UI's
+   * minimum size.
+   *
+   * @return the preferred size for the JEditorPane
    */
   public Dimension getPreferredSize()
   {
-    return super.getPreferredSize();
+    Dimension pref = super.getPreferredSize();
+    Container parent = getParent();
+    if (parent instanceof JViewport)
+      {
+        JViewport vp = (JViewport) getParent();
+        TextUI ui = getUI();
+        Dimension min = null;
+        if (! getScrollableTracksViewportWidth())
+          {
+            min = ui.getMinimumSize(this);
+            int vpWidth = vp.getWidth();
+            if (vpWidth != 0 && vpWidth < min.width)
+              pref.width = min.width;
+          }
+        if (! getScrollableTracksViewportHeight())
+          {
+            if (min == null)
+              min = ui.getMinimumSize(this);
+            int vpHeight = vp.getHeight();
+            if (vpHeight != 0 && vpHeight < min.height)
+              pref.height = min.height;
+          }
+      }
+    return pref;
   }
 
+  /**
+   * Returns <code>true</code> when a Viewport should force the height of
+   * this component to match the viewport height. This is implemented to return
+   * <code>true</code> when  the parent is an instance of JViewport and
+   * the viewport height > the UI's minimum height.
+   *
+   * @return <code>true</code> when a Viewport should force the height of
+   *         this component to match the viewport height
+   */
   public boolean getScrollableTracksViewportHeight()
   {
-  /*  Container parent = getParent();
-    return (parent instanceof JViewport &&
-        parent.isValid());*/
-    return isValid();
+    // Tests show that this returns true when the parent is a JViewport
+    // and has a height > minimum UI height.
+    Container parent = getParent();
+    int height = parent.getHeight();
+    TextUI ui = getUI();
+    return parent instanceof JViewport
+           && height >= ui.getMinimumSize(this).height
+           && height <= ui.getMaximumSize(this).height;
   }
 
+  /**
+   * Returns <code>true</code> when a Viewport should force the width of
+   * this component to match the viewport width. This is implemented to return
+   * <code>true</code> when  the parent is an instance of JViewport and
+   * the viewport width > the UI's minimum width.
+   *
+   * @return <code>true</code> when a Viewport should force the width of
+   *         this component to match the viewport width
+   */
   public boolean getScrollableTracksViewportWidth()
   {
-    /*Container parent = getParent();
-    return (parent instanceof JViewport &&
-        parent.isValid());*/
-    return isValid();
+    // Tests show that this returns true when the parent is a JViewport
+    // and has a width > minimum UI width.
+    Container parent = getParent();
+    return parent != null && parent instanceof JViewport
+           && parent.getWidth() > getUI().getMinimumSize(this).width;
   }
 
   public URL getPage()
   {
-    return page;
+    return loader != null ? loader.page : null;
   }
 
   protected InputStream getStream(URL page)
     throws IOException
   {
-    return page.openStream();
+    URLConnection conn = page.openConnection();
+    // Try to detect the content type of the stream data.
+    String type = conn.getContentType();
+    if (type != null)
+      setContentType(type);
+    InputStream stream = conn.getInputStream();
+    return new BufferedInputStream(stream);
   }
 
   public String getText()
@@ -651,10 +982,12 @@ public class JEditorPane extends JTextComponent
     EditorKit kit = getEditorKit();
     if (kit instanceof HTMLEditorKit && desc instanceof HTMLDocument)
       {
-        Document doc = (Document) desc;
+        HTMLDocument doc = (HTMLDocument) desc;
+        setDocument(doc);
         try
           {
-            kit.read(in, doc, 0);
+            InputStreamReader reader = new InputStreamReader(in);
+            kit.read(reader, doc, 0);
           }
         catch (BadLocationException ex)
           {
@@ -669,12 +1002,18 @@ public class JEditorPane extends JTextComponent
   }
 
   /**
-   * Establishes the default bindings of type to classname. 
+   * Establishes a binding between type and classname.  This enables
+   * us to create an EditorKit later for the given content type.
+   * 
+   * @param type the content type
+   * @param classname the name of the class that is associated with this 
+   * content type
    */
   public static void registerEditorKitForContentType(String type,
                                                      String classname)
   {
-    // TODO: Implement this properly.
+    registerEditorKitForContentType(type, classname,
+                               Thread.currentThread().getContextClassLoader());
   }
 
   /**
@@ -684,7 +1023,7 @@ public class JEditorPane extends JTextComponent
                                                      String classname,
                                                      ClassLoader loader)
   {
-    // TODO: Implement this properly.
+    registerMap.put(type, new EditorKitMapping(classname, loader));
   }
 
   /**
@@ -694,6 +1033,7 @@ public class JEditorPane extends JTextComponent
   public void replaceSelection(String content)
   {
     // TODO: Implement this properly.
+    super.replaceSelection(content);
   }
 
   /**
@@ -707,6 +1047,13 @@ public class JEditorPane extends JTextComponent
 
   public final void setContentType(String type)
   {
+    // Strip off content type parameters.
+    int paramIndex = type.indexOf(';');
+    if (paramIndex > -1)
+      {
+        // TODO: Handle character encoding.
+        type = type.substring(0, paramIndex).trim();
+      }
     if (editorKit != null
 	&& editorKit.getContentType().equals(type))
       return;
@@ -741,9 +1088,14 @@ public class JEditorPane extends JTextComponent
     accessibleContext = null;
   }
 
+  /**
+   * Explicitly sets an EditorKit to be used for the given content type.
+   * @param type the content type
+   * @param k the EditorKit to use for the given content type
+   */
   public void setEditorKitForContentType(String type, EditorKit k)
   {
-    // FIXME: editorKitCache.put(type, kit);
+    editorMap.put(type, k);
   }
 
   /**
@@ -762,20 +1114,78 @@ public class JEditorPane extends JTextComponent
     if (page == null)
       throw new IOException("invalid url");
 
-    try
+    URL old = getPage();
+    // Only reload if the URL doesn't point to the same file.
+    // This is not the same as equals because there might be different
+    // URLs on the same file with different anchors.
+    if (old == null || ! old.sameFile(page))
       {
-	this.page = page;
-	getEditorKit().read(page.openStream(), getDocument(), 0);
-      }
-    catch (BadLocationException e)
-      {
-	// Ignored. '0' is always a valid offset.
+        InputStream in = getStream(page);
+        if (editorKit != null)
+          {
+            Document doc = editorKit.createDefaultDocument();
+            doc.putProperty(Document.StreamDescriptionProperty, page);
+
+            if (loader != null)
+              loader.cancel();
+            loader = new PageLoader(doc, in, old, page);
+
+            int prio = -1;
+            if (doc instanceof AbstractDocument)
+              {
+                AbstractDocument aDoc = (AbstractDocument) doc;
+                prio = aDoc.getAsynchronousLoadPriority();
+              }
+            if (prio >= 0)
+              {
+                // Load asynchronously.
+                setDocument(doc);
+                Thread loadThread = new Thread(loader,
+                                               "JEditorPane.PageLoader");
+                loadThread.setDaemon(true);
+                loadThread.setPriority(prio);
+                loadThread.start();
+              }
+            else
+              {
+                // Load synchronously.
+                loader.run();
+                setDocument(doc);
+              }
+          }
       }
   }
 
+  /**
+   * Sets the text of the JEditorPane.  The argument <code>t</code>
+   * is expected to be in the format of the current EditorKit.  This removes
+   * the content of the current document and uses the EditorKit to read in the
+   * new text.  This allows the EditorKit to handle the String rather than just
+   * inserting in plain text.
+   * 
+   * @param t the text to display in this JEditorPane
+   */
   public void setText(String t)
   {
-    super.setText(t);
+    try
+    {
+      // Remove the current content.
+      Document doc = getDocument();
+      doc.remove(0, doc.getLength());
+      if (t == null || t.equals(""))
+        return;
+      
+      // Let the EditorKit read the text into the Document.
+      getEditorKit().read(new StringReader(t), doc, 0);
+    }
+    catch (BadLocationException ble)
+    {
+      // TODO: Don't know what to do here.
+    }
+    catch (IOException ioe)
+    {
+      // TODO: Don't know what to do here.
+    }
   }
 
   /**
